% ========================================
% TWO TIME SCALES LEAST SQUARES
% ========================================

% The user's parameters here are the poles
% of the prefilters, ts and tf.
% Put your data into the next two lines. 
ts = 1/50; % low pass filter's double pole
tf = 1/100; % high pass filter's double pole

% ----------------------------------------
% slowscale and fastscale are the inverse 
% of the time step that will be used to 
% compute the finite difference operator
% that will define the slow and fast 
% excitation matrices.
% This implies a reparametrization.

% The following choice is a rule of thumb 
% that can be modified.
slowscale = min(max(abs(Pp)),abs(Zp))/4
fastscale = min(abs(Pp))/4

% Filter the signals first using second 
% order rational filters. They have no 
% zero, a double pole, and a static
% (resp. HF) gain equal to 1. 
% Of course, you may change that.

% ---------------------------------------
% slow prefiltering
disp('slow prefiltering...')
zfs = [];
pfs = [-ts -ts];
kfs = ts^2;
ys = dofilter(zfs,pfs,kfs,y,1);
us = dofilter(zfs,pfs,kfs,u,1);

% ---------------------------------------
% fast prefiltering
disp('fast prefiltering...')
zff = [];
pff = [-tf -tf];
kff = tf^2;
yf = y-dofilter(zff,pff,kff,y,1);
uf = u-dofilter(zff,pff,kff,u,1);

% ---------------------------------------
% normalize the signals
dus = max(abs(us));
dys = max(abs(ys));
us = us/dus;
ys = ys/dys;
norms = dys/dus;

duf = max(abs(uf));
dyf = max(abs(yf));
uf = uf/duf;
yf = yf/dyf;
normf = dyf/duf;

% ---------------------------------------
% compute excitations using time scaled 
% finite difference operators.
% The last parameter of excDelta is the 
% time step; see below for details.
disp('Computing excitations')
Qstemp = excDelta(ys,us,1,1,1/slowscale); 
% notice the time rescaling
Qs = Qstemp(2:4,2:4);
Rs = Qstemp(2:4,1);
Qftemp = excDelta(yf,uf,0,1,1/fastscale); 
% notice the time rescaling
% Some reordering needed here because of 
% special form
Qf = Qftemp([1 3],[1 3]); 
Rf = Qftemp([1 3],2);

% ---------------------------------------
% Uzawa algorithm to solve the optimization
% problem. Each low level minimization is 
% a modified LS. The constraint is: 
% sp(2)*norms - fp(2)*normf = 0
% with sp = slow parameters and 
% fp = fast parameters, that is,
% slow HF gain = fast static gain.
% The cost is: 1/2 (sp'*Qs*sp - 2 sp'*Rs) 
% + 1/2 (fp'*Qf*fp - 2 fp'*Rf),
% that is, the sum of the slow and fast 
% prediction error costs (up to a constant).

% Initialization
sp=zeros(1,3); % slow parameters
fp=zeros(1,2); % fast parameters
tol = sqrt(eps); % stop test value
p = 0; % initial price for the constraint
cons = 1+tol; % initial constraint level
% the algorithm stops when cons is 
% smaller than tol.

% compute a nice step using the theory 
% for the linear quadratic case
lambda1 = min(eig(Qs));
lambda2 = min(eig(Qf));
% cost convexity
lambda = min([lambda1 lambda2]); 
% square of constraint's gradient
d2 = norms^2 + normf^2; 
step = lambda/(4*d2); % this is it
% Precompute inverses
Qs1=inv(Qs);
Qf1=inv(Qf);
% main loop here
while abs(cons) > tol
% low level minimization of the Lagrangian; 
% solution has a known form,
% so we compute it directly
sp = Qs1*(Rs-[0;p*norms;0]);
fp = Qf1*(Rf+[0;p*normf]);
% compute constraint
cons=sp(2)*norms-fp(2)*normf;
% update price with a gradient step (upwards)
p = p+(step*cons);
end;
% end of loop

% ---------------------------------------
% Recover parameters
% Go back to the advance operator
ns = [sp(2),sp(3)-sp(2)]*norms*(ny/nu);
ds = [1, -1-sp(1)];
% convert to time constants parameters
[As,Bs,Cs,Ds]=tf2ss(ns,ds);
[As,Bs]=d2c(As,Bs,1); % rescale later
[zs,ps,ks]=ss2zp(As,Bs,Cs,Ds);
[zs,ps,ks]=zp2tc(zs,ps,ks); 
% Rescale slow parameters
zs = zs*slowscale;
ps = ps*slowscale;
% get fast time constant directly
pf = log(1+(1/(fp(1)*fastscale)));
pf = 1/pf; 
% compute error
parms2 = [zs , pf , ps , ks]';
err2 = (parms2-[Zp;Pp;Kp])./[Zp;Pp;Kp];
disp('    ref.      ident.    rel. err.')
disp([[Zp;Pp;Kp],parms2,err2])
